const coral = @import("coral");

const tokens = @import("./tokens.zig");

name: []const coral.io.Byte,
allocator: coral.io.Allocator,
arena: coral.arena.Stacking,
statements: Statement.List,
error_buffer: coral.list.ByteStack,
tokenizer: tokens.Tokenizer,

pub const Expression = union (enum) {
	nil_literal,
	true_literal,
	false_literal,
	number_literal: []const coral.io.Byte,
	string_literal: []const coral.io.Byte,
	symbol_literal: []const coral.io.Byte,
	table_literal: TableLiteral,
	grouped_expression: *Expression,
	resolve_local: []const coral.io.Byte,

	get_field: struct {
		object_expression: *Expression,
		identifier: []const coral.io.Byte,
	},

	set_field: struct {
		object_expression: *Expression,
		identifier: []const coral.io.Byte,
		value_expression: *Expression,
	},

	call_system: struct {
		identifier: []const coral.io.Byte,
		argument_expressions: List,
	},

	binary_operation: struct {
		operator: BinaryOperator,
		lhs_expression: *Expression,
		rhs_expression: *Expression,
	},

	unary_operation: struct {
		operator: UnaryOperator,
		expression: *Expression,
	},

	pub const BinaryOperator = enum {
		addition,
		subtraction,
		multiplication,
		divsion,
		equals_comparison,
		greater_than_comparison,
		greater_equals_comparison,
		less_than_comparison,
		less_equals_comparison,

		fn token(self: BinaryOperator) tokens.Token {
			return switch (self) {
				.addition => .symbol_plus,
				.subtraction => .symbol_minus,
				.multiplication => .symbol_asterisk,
				.divsion => .symbol_forward_slash,
				.equals_comparison => .symbol_double_equals,
				.greater_than_comparison => .symbol_greater_than,
				.greater_equals_comparison => .symbol_greater_equals,
				.less_than_comparison => .symbol_less_than,
				.less_equals_comparison => .symbol_less_equals,
			};
		}
	};

	pub const TableLiteral = coral.list.Stack(struct {
		key_expression: Expression,
		value_expression: Expression,
	});

	pub const List = coral.list.Stack(Expression);
};

const ExpressionParser = fn (self: *Self) ParseError!Expression;

pub const ParseError = error {
	OutOfMemory,
	BadSyntax,
};

const Self = @This();

pub const Statement = union (enum) {
	return_nothing,
	return_expression: Expression,

	assign_local: struct {
		identifier: []const coral.io.Byte,
		expression: Expression,
	},

	call_system: struct {
		identifier: []const coral.io.Byte,
		argument_expressions: Expression.List,
	},

	const List = coral.list.Stack(Statement);
};

const UnaryOperator = enum {
	boolean_negation,
	numeric_negation,
};

fn binary_operation_parser(
	comptime parse_next: ExpressionParser,
	comptime operators: []const Expression.BinaryOperator) ExpressionParser {

	const BinaryOperationParser = struct {
		fn parse(self: *Self) ParseError!Expression {
			const allocator = self.arena.as_allocator();
			var expression = try parse_next(self);

			inline for (operators) |operator| {
				const token = comptime operator.token();

				if (self.tokenizer.token == coral.io.tag_of(token)) {
					self.tokenizer.step();

					if (self.tokenizer.token == .end) {
						return self.report("expected other half of expression after `" ++ comptime token.text() ++ "`");
					}

					expression = .{
						.binary_operation = .{
							.operator = operator,
							.lhs_expression = try coral.io.allocate_one(allocator, expression),
							.rhs_expression = try coral.io.allocate_one(allocator, try parse_next(self)),
						},
					};
				}
			}

			return expression;
		}
	};

	return BinaryOperationParser.parse;
}

pub fn error_message(self: Self) []const coral.io.Byte {
	return self.error_buffer.values;
}

pub fn free(self: *Self) void {
	self.arena.free();
	self.statements.free();
	self.error_buffer.free();
}

pub fn make(allocator: coral.io.Allocator, ast_name: []const coral.io.Byte) Self {
	return Self{
		.arena = coral.arena.Stacking.make(allocator, 4096),
		.error_buffer = coral.list.ByteStack.make(allocator),
		.statements = Statement.List.make(allocator),
		.tokenizer = .{.source = ""},
		.allocator = allocator,
		.name = ast_name,
	};
}

fn report(self: *Self, message: []const coral.io.Byte) ParseError {
	coral.utf8.print_formatted(coral.list.stack_as_writer(&self.error_buffer), "{name}@{line}: {message}", .{
		.name = self.name,
		.line = self.tokenizer.lines_stepped,
		.message = message,
	}) catch return error.OutOfMemory;

	return error.BadSyntax;
}

pub fn list_statements(self: Self) []const Statement {
	return self.statements.values;
}

pub fn parse(self: *Self, data: []const coral.io.Byte) ParseError!void {
	self.tokenizer = .{.source = data};

	const allocator = self.arena.as_allocator();
	var has_returned = false;

	self.tokenizer.skip_newlines();

	while (true) {
		switch (self.tokenizer.token) {
			.end => return,

			.keyword_return => {
				if (has_returned) {
					return self.report("multiple returns in function scope but expected only one");
				}

				try self.statements.push_one(get_statement: {
					self.tokenizer.step();

					if (self.tokenizer.token != .end and self.tokenizer.token != .newline) {
						break: get_statement .{.return_expression = try self.parse_expression()};
					}

					if (self.tokenizer.token != .end and self.tokenizer.token != .newline) {
						return self.report("expected end or newline after return statement");
					}

					break: get_statement .return_nothing;
				});

				has_returned = true;
			},

			.identifier => |identifier| {
				self.tokenizer.step();

				switch (self.tokenizer.token) {
					.end, .newline => return self.report("statement has no effect"),

					.symbol_equals => {
						self.tokenizer.step();

						if (self.tokenizer.token == .end) {
							return self.report("expected expression after `=`");
						}

						try self.statements.push_one(.{
							.assign_local = .{
								.expression = try self.parse_expression(),
								.identifier = identifier,
							},
						});
					},

					else => return self.report("expected `=` after local"),
				}
			},

			.special_identifier => |identifier| {
				self.tokenizer.step();

				switch (self.tokenizer.token) {
					.end, .newline => return self.report("system call is missing arguments"),

					.symbol_paren_left => {
						self.tokenizer.step();

						var expressions_list = Expression.List.make(allocator);

						while (true) {
							if (self.tokenizer.token == .symbol_paren_right) {
								break;
							}

							try expressions_list.push_one(try self.parse_expression());

							switch (self.tokenizer.token) {
								.symbol_comma => continue,
								.symbol_paren_right => break,
								else => return self.report("expected `)` or argument after `(`"),
							}
						}

						self.tokenizer.skip_newlines();

						try self.statements.push_one(.{
							.call_system = .{
								.argument_expressions = expressions_list,
								.identifier = identifier,
							},
						});
					},

					else => return self.report("expected `=` after local"),
				}
			},

			else => return self.report("invalid statement"),
		}
	}
}

const parse_comparison = binary_operation_parser(parse_term, &.{
	.greater_than_comparison,
	.greater_equals_comparison,
	.less_than_comparison,
	.less_equals_comparison
});

const parse_equality = binary_operation_parser(parse_comparison, &.{
	.equals_comparison,
});

const parse_expression = binary_operation_parser(parse_equality, &.{
	.addition,
	.subtraction,
});

fn parse_factor(self: *Self) ParseError!Expression {
	const allocator = self.arena.as_allocator();

	var expression = @as(Expression, get: {
		switch (self.tokenizer.token) {
			.symbol_paren_left => {
				self.tokenizer.skip_newlines();

				if (self.tokenizer.token == .end) {
					return self.report("expected an expression after `(`");
				}

				const expression = try self.parse_expression();

				if (self.tokenizer.token != .symbol_paren_right) {
					return self.report("expected a closing `)` after expression");
				}

				self.tokenizer.skip_newlines();

				break: get Expression{.grouped_expression = try coral.io.allocate_one(allocator, expression)};
			},

			.keyword_nil => {
				self.tokenizer.skip_newlines();

				break: get .nil_literal;
			},

			.keyword_true => {
				self.tokenizer.skip_newlines();

				break: get .true_literal;
			},

			.keyword_false => {
				self.tokenizer.skip_newlines();

				break: get .false_literal;
			},

			.number => |value| {
				self.tokenizer.skip_newlines();

				break: get .{.number_literal = value};
			},

			.string => |value| {
				self.tokenizer.skip_newlines();

				break: get .{.string_literal = value};
			},

			.special_identifier => |identifier| {
				self.tokenizer.skip_newlines();

				var expression_list = Expression.List.make(allocator);

				while (true) {
					switch (self.tokenizer.token) {
						.end => return self.report("expected expression or `)` after `(`"),

						.symbol_paren_right => {
							self.tokenizer.skip_newlines();

							break: get .{
								.call_system = .{
									.identifier = identifier,
									.argument_expressions = expression_list,
								},
							};
						},

						else => {
							try expression_list.push_one(try self.parse_expression());

							switch (self.tokenizer.token) {
								.end => return self.report("expected `,` or `)` after argument"),
								.symbol_comma => continue,

								.symbol_paren_right => {
									self.tokenizer.skip_newlines();

									break: get .{
										.call_system = .{
											.identifier = identifier,
											.argument_expressions = expression_list,
										},
									};
								},

								else => return self.report("expected `,` or `)` after argument"),
							}
						},
					}
				}
			},

			.identifier => |identifier| {
				self.tokenizer.skip_newlines();

				break: get .{.resolve_local = identifier};
			},

			.symbol_brace_left => {
				var table_literal = Expression.TableLiteral.make(allocator);

				self.tokenizer.skip_newlines();

				while (true) {
					switch (self.tokenizer.token) {
						.symbol_brace_right => {
							self.tokenizer.skip_newlines();

							break: get .{.table_literal = table_literal};
						},

						.symbol_bracket_left => {
							self.tokenizer.skip_newlines();

							if (self.tokenizer.token != .symbol_equals) {
								return self.report("expected expression after identifier");
							}
						},

						.symbol_period => {
							self.tokenizer.step();

							const identifier = switch (self.tokenizer.token) {
								.identifier => |identifier| identifier,
								else => return self.report("expected identifier after `.`"),
							};

							self.tokenizer.skip_newlines();

							if (self.tokenizer.token != .symbol_equals) {
								return self.report("expected `=` after key");
							}

							self.tokenizer.skip_newlines();

							if (self.tokenizer.token == .end) {
								return self.report("unexpected end after `=`");
							}

							try table_literal.push_one(.{
								.value_expression = try self.parse_expression(),
								.key_expression = .{.symbol_literal = identifier},
							});

							switch (self.tokenizer.token) {
								.symbol_comma => self.tokenizer.skip_newlines(),

								.symbol_brace_right => {
									self.tokenizer.skip_newlines();

									break: get .{.table_literal = table_literal};
								},

								else => return self.report("expected `,` or `}` after expression"),
							}
						},

						else => return self.report("expected `}` or fields in table literal"),
					}
				}
			},

			.symbol_minus => {
				self.tokenizer.skip_newlines();

				if (self.tokenizer.token == .end) {
					return self.report("expected expression after numeric negation (`-`)");
				}

				break: get .{
					.unary_operation = .{
						.expression = try coral.io.allocate_one(allocator, try self.parse_factor()),
						.operator = .numeric_negation,
					},
				};
			},

			.symbol_bang => {
				self.tokenizer.skip_newlines();

				if (self.tokenizer.token == .end) {
					return self.report("expected expression after boolean negation (`!`)");
				}

				break: get .{
					.unary_operation = .{
						.expression = try coral.io.allocate_one(allocator, try self.parse_factor()),
						.operator = .boolean_negation,
					},
				};
			},

			else => return self.report("unexpected token in expression"),
		}
	});

	while (self.tokenizer.token == .symbol_period) {
		self.tokenizer.skip_newlines();

		const identifier = switch (self.tokenizer.token) {
			.identifier => |identifier| identifier,
			else => return self.report("expected identifier after `.`"),
		};

		self.tokenizer.skip_newlines();

		expression = switch (self.tokenizer.token) {
			.symbol_equals => .{
				.set_field = .{
					.value_expression = try coral.io.allocate_one(allocator, try self.parse_expression()),
					.object_expression = try coral.io.allocate_one(allocator, expression),
					.identifier = identifier,
				},
			},

			else => .{
				.get_field = .{
					.object_expression = try coral.io.allocate_one(allocator, expression),
					.identifier = identifier,
				},
			},
		};
	}

	return expression;
}

const parse_term = binary_operation_parser(parse_factor, &.{
	.multiplication,
	.divsion,
});
