const coral = @import("coral");

const kym = @import("../kym.zig");

associative: AssociativeTable,
contiguous: ContiguousList,

const AssociativeTable = coral.map.Table(*kym.RuntimeRef, *kym.RuntimeRef, struct {
	env: *kym.RuntimeEnv,

	pub fn hash(self: @This(), key_ref: *kym.RuntimeRef) usize {
		return kym.hash(self.env, key_ref);
	}

	pub fn equals(self: @This(), key_ref_a: *kym.RuntimeRef, key_ref_b: *kym.RuntimeRef) bool {
		return kym.test_equality(self.env, key_ref_a, key_ref_b);
	}
});

const ContiguousList = coral.list.Stack(?*kym.RuntimeRef);

const Self = @This();

pub fn new(env: *kym.RuntimeEnv) kym.RuntimeError!*kym.RuntimeRef {
	var self = Self{
		.associative = AssociativeTable.make(env.allocator, .{
			.env = env,
		}),

		.contiguous = ContiguousList.make(env.allocator),
	};

	errdefer {
		self.associative.free();
		self.contiguous.free();
	}

	return try env.new_dynamic(coral.io.bytes_of(&self), &typeinfo);
}

const typeinfo = kym.Typeinfo{
	.name = "table",
	.clean = typeinfo_clean,
	.get = typeinfo_get,
	.set = typeinfo_set,
};

fn typeinfo_clean(method: kym.Method) void {
	const table = @as(*Self, @ptrCast(@alignCast(method.userdata.ptr)));

	{
		var field_iterable = table.associative.as_iterable();

		while (field_iterable.next()) |entry| {
			method.env.discard(entry.key);
			method.env.discard(entry.value);
		}
	}

	table.associative.free();

	while (table.contiguous.pop()) |ref| {
		method.env.discard(ref);
	}

	table.contiguous.free();
}

fn typeinfo_get(method: kym.Method, index_ref: *const kym.RuntimeRef) kym.RuntimeError!?*kym.RuntimeRef {
	const table = @as(*Self, @ptrCast(@alignCast(method.userdata.ptr)));

	switch (method.env.unbox(index_ref)) {
		.symbol, .string, .float => {
			const mutable_index_ref = method.env.acquire(index_ref);

			defer method.env.discard(mutable_index_ref);

			if (table.associative.lookup(mutable_index_ref)) |value_ref| {
				return method.env.acquire(value_ref);
			}
		},

		.fixed => |fixed| {
			if (fixed < 0) {
				// TODO: Negative indexing.
				unreachable;
			}

			if (fixed < table.contiguous.values.len) {
				return method.env.acquire(table.contiguous.values[@intCast(fixed)] orelse return null);
			}

			const mutable_index_ref = method.env.acquire(index_ref);

			defer method.env.discard(mutable_index_ref);

			if (table.associative.lookup(mutable_index_ref)) |value_ref| {
				return method.env.acquire(value_ref);
			}
		},

		else => return method.env.raise(error.TypeMismatch, "expected symbol, string, or number index"),
	}

	return null;
}

fn typeinfo_set(method: kym.Method, index_ref: *const kym.RuntimeRef, value_ref: ?*const kym.RuntimeRef) kym.RuntimeError!void {
	const table = @as(*Self, @ptrCast(@alignCast(method.userdata.ptr)));

	switch (method.env.unbox(index_ref)) {
		.symbol, .string, .float => {
			const acquired_index_ref = method.env.acquire(index_ref);

			errdefer method.env.discard(acquired_index_ref);

			if (value_ref) |ref| {
				const acquired_ref = method.env.acquire(ref);

				errdefer method.env.discard(acquired_ref);

				if (try table.associative.replace(acquired_index_ref, acquired_ref)) |replaced| {
					method.env.discard(replaced.key);
					method.env.discard(replaced.value);
				}
			} else if (table.associative.remove(acquired_index_ref)) |removed| {
				method.env.discard(removed.key);
				method.env.discard(removed.value);
			}
		},

		.fixed => |fixed| {
			if (fixed < 0) {
				// TODO: Negative indexing.
				unreachable;
			}

			if (fixed < table.contiguous.values.len) {
				const contiguous_value = &table.contiguous.values[@intCast(fixed)];

				method.env.discard(contiguous_value.*);

				contiguous_value.* = if (value_ref) |ref| method.env.acquire(ref) else null;

				return;
			}

			if (fixed == table.contiguous.values.len) {
				if (value_ref) |ref| {
					const acquired_ref = method.env.acquire(ref);

					errdefer method.env.discard(acquired_ref);

					try table.contiguous.push_one(acquired_ref);
				} else {
					try table.contiguous.push_one(null);
				}

				return;
			}

			const acquired_index_ref = method.env.acquire(index_ref);

			errdefer method.env.discard(acquired_index_ref);

			if (value_ref) |ref| {
				const acquired_ref = method.env.acquire(ref);

				errdefer method.env.discard(acquired_ref);

				if (try table.associative.replace(acquired_index_ref, acquired_ref)) |replaced| {
					method.env.discard(replaced.key);
					method.env.discard(replaced.value);
				}
			} else if (table.associative.remove(acquired_index_ref)) |removed| {
				method.env.discard(removed.key);
				method.env.discard(removed.value);
			}
		},

		else => return method.env.raise(error.TypeMismatch, "expected symbol, string, or number index"),
	}
}
